#define __ENABLE_WSTRING 1
#include <cstdio>
#include <string>
#include <unistd.h>
#include <csignal>
#include <sys/wait.h>
#include <vector>
#include <cctype>
#include <map>
#include <set>
#include <list>
#include <cerrno>

#include <iconv.h>
#include "autoptr"

#define DEBUG 0

using namespace std;

//#define wstring ucs4string
typedef wchar_t ucs4;
//typedef unsigned int ucs4;
//typedef basic_string<ucs4> wstring;

static const ucs4 ilseq = 0xFFFD;

#if 1
/* libstdc++ 2 way */
static const wstring emptywstring;
#define CLEARSTR(x) x=emptywstring
#else
/* libstdc++ 3 way */
#define CLEARSTR(x) x.clear()
#endif

#if 0
static bool operator==(const wstring &s1, const char *s2)
{
    for(unsigned a=0; a<s1.size(); ++a, ++s2)
    {
        if(!*s2 || s1[a] != (unsigned char)*s2)return false;
    }
    return *s2 ? false : true;
}
#endif
static void operator+= (wstring &s1, const char *s2)
{
    while(*s2)s1 += *s2++;
}
static bool IsEqual(const wstring &s1, const char *s2)
{
    for(unsigned a=0; a<s1.size(); ++a, ++s2)
    {
        if(!*s2)return false;
        if(s1[a] >= 0x100)return false;
        char c1 = (char)s1[a];
        if(toupper(c1) != toupper(*s2))return false;
    }
    return !*s2;
}

static const string Stringify(const wstring &s)
{
	string result(s.size(), '?');
	for(unsigned a=0; a<s.size(); ++a)
		if(s[a] < 0x100) result[a] = (char)s[a];
	return result;
}

static ucs4 Getc(FILE *fp)
{
    ucs4 p; fread(&p, 1, 4, fp);
    return p;
}

static const struct { const char *ent; ucs4 ch; } EnTab[] = {
#include "entities.h"
};
#define ENTITYCOUNT (sizeof(EnTab) / sizeof(EnTab[0]))
static ucs4 FindEntity(const wstring &ws)
{
    string s(ws.size(), ' ');
    for(unsigned a=0; a<ws.size(); ++a)
        s[a] = (char)ws[a];
    
    unsigned eka=0, vika=ENTITYCOUNT-1;
    for(;;)
    {
        unsigned n = (eka+vika)/2;
        if(s == EnTab[n].ent)return EnTab[n].ch;
        if(eka >= vika)break;
        if(s > EnTab[n].ent) { eka=n+1; continue; }
        if(s < EnTab[n].ent) { vika=n; continue; }
    }
    return ilseq;
}

#undef putc
#undef puts

static unsigned FixedStyleScript = 0;
static unsigned ParamViolations  = 0;
static const char *const midset = (*(const short *)"\1\0\0\0" == 1) ? "UCS-4LE" : "UCS-4BE";
static bool lossless = true;
static bool usehex = false;
static bool strict = false;
static int verbose = 1;

class Page
{
    class Dumper
    {
        mutable iconv_t converter;
        mutable iconv_t tester;
        
        bool OpenConv(iconv_t &conv, const char *set1, const char *set2)
        {
            conv = iconv_open(set1, set2);
            if(conv == (iconv_t)-1)
            {
                fprintf(stderr, "iconv_open failed to create '%s' to '%s' converter. Aborting.\n",
                    set1, set2);
                return true;
            }
            return false;
        }
    public:
        string charset;
        
        Dumper() : charset(midset)
        {
            if(OpenConv(converter, charset.c_str(), midset)
            || OpenConv(tester,    charset.c_str(), midset))
                exit(EINVAL);
        }
        ~Dumper()
        {
            iconv_close(converter);
            iconv_close(tester);
        }
        
        void SetSet(const char *setname)
        {
            if(verbose >= 1)
            {
                fprintf(stderr, "Recoding %s to output (%s)\n", midset, setname);
                fflush(stderr);
            }
            iconv_close(converter);
            iconv_close(tester);
            charset = setname;
            if(OpenConv(converter, setname, midset)
            || OpenConv(tester,    setname, midset))
                exit(EINVAL);
        }
        
        void putc(ucs4 p) const
        {
            wstring tmp;
            tmp += p;
            puts(tmp);
        }
        
        void puts(const wstring &s) const
        {
            char *input = (char *) (const_cast<ucs4 *> (s.data()));
            size_t left = s.size() * sizeof(ucs4);
            while(left > 0)
            {
                char OutBuf[4096], *outptr = OutBuf;
                unsigned outsize = sizeof OutBuf;
            #if DEBUG
                fprintf(stderr, "P1:Converting %u bytes to %u bytes space\n", left, outsize);
                unsigned bytesread = left, converted = outsize;
            #endif
                size_t retval = iconv(converter, &input, &left, &outptr, &outsize);
            #if DEBUG
                bytesread -= left; converted -= outsize;
                fprintf(stderr, "%u bytes read, %u bytes left, %u bytes generated, %u bytes  space left\n", bytesread, left, converted, outsize);
            #endif
                fwrite(OutBuf, 1, outptr-OutBuf, stdout);
                if(retval == (size_t)-1)
                {
        #if DEBUG
                    perror("iconv");
        #endif
                    if(errno == E2BIG)
                    {
                        continue;
                    }
                    if(errno == EILSEQ)
                    {
                        input += sizeof(ucs4);
                        left -= sizeof(ucs4);
                        putchar('?');
                    }
                    if(errno == EINVAL)
                    {
                        /* Got partial byte */
                    }
                }
            }
        }
        
        bool isok(ucs4 p) const
        {
            char OutBuf[256], *outptr = OutBuf, *tmp = (char *)&p;
            unsigned outsize = sizeof OutBuf;
            unsigned insize = sizeof(p);
            size_t retval = iconv(tester, &tmp, &insize, &outptr, &outsize);
            if(retval == (size_t)-1)return false;
            return true;
        }
    } Dumper;
    
    bool CanDump(ucs4 p) const
    {
        return Dumper.isok(p);
    }
    
    mutable char underquote;

    wstring htmlencode(const wstring &s) const;
    wstring htmldecode(const wstring &s) const;
    
    struct Tag
    {
        typedef map<wstring, wstring>::iterator iterator;
        typedef map<wstring, wstring>::const_iterator const_iterator;
        wstring Name;
        map<wstring, wstring> Params;
        set<wstring> Keys;
        list<wstring> Comments;
        
        iterator begin() { return Params.begin(); }
        iterator end() { return Params.end(); }
        const_iterator begin() const { return Params.begin(); }
        const_iterator end() const { return Params.end(); }

        iterator Ifind(const char *s)
        {
            for(iterator i=Params.begin(); ; ++i)
                if(i == Params.end()
                || IsEqual(i->first, s))
                    return i;
        }
        const_iterator Ifind(const char *s) const
        {
            for(const_iterator i=Params.begin(); ; ++i)
                if(i == Params.end()
                || IsEqual(i->first, s))
                    return i;
        }
        
        void SetParam(const wstring &name,
                      const wstring &value)
        {
            Params[name] = value;
        }
        void SetParam(const wstring &name)
        {
            Keys.insert(name);
        }
        
        void clear()
        {
            CLEARSTR(Name);
            Params.clear();
            Keys.clear();
            Comments.clear();
        }
        bool Is(const char *t) const
        {
            return IsEqual(Name, t);
        }
        bool Is(const wstring &s) const
        {
            if(s.size() != Name.size()) return false;
            for(unsigned a=0; a<s.size(); ++a)
            {
                if(s[a] == Name[a]
                || (s[a] <= 0x100 && Name[a] <= 0x100
                && toupper((char)s[a]) == toupper((char)Name[a]))
                  ) continue;
                return false;
            }
            return true;
        }
    };
    
    void Putc(ucs4 p) const
    {
        Dumper.putc(p);
    }
    void Dump(const wstring &s) const
    {
        Dumper.puts(s);
    }
    void DumpHTML(const wstring &s) const
    {
        Dump(htmlencode(s));
    }
    void DumpHTML(const wstring &s, char quotetype) const
    {
        underquote = quotetype;
        DumpHTML(s);
        underquote = 0;
    }
    void DumpTag(const Tag &tag) const
    {
        Putc('<');
        Dump(tag.Name);
        Tag::const_iterator i;

        /* FIXME: Parameters are shuffled
                  (Doesn't break anything,
                   but may cause curiosity)
         */
        for(i=tag.Params.begin(); i!=tag.Params.end(); ++i)
        {
            Putc(' ');
            Dump(i->first);
            Putc('=');
            const wstring &s = i->second;
            bool needquotes = false;
            
            unsigned countq2 = 0;
            unsigned countq1 = 0;
            
            for(unsigned a=0; a<s.size(); ++a)
            {
                if(s[a] >= 'A' && s[a] <= 'Z')continue;
                if(s[a] >= 'a' && s[a] <= 'z')continue;
                if(s[a] >= '0' && s[a] <= '9')continue;
                if(s[a] == '-' || s[a] == '.'
                || s[a] == '_' || s[a] == ':')continue;
                if(s[a] == '"') ++countq2;
                if(s[a] == '\'') ++countq1;
                needquotes = true;
                break;
            }
            
            if(!s.size()) needquotes = true;
            
            if(needquotes)
            {
                char quotetype = countq2 <= countq1 ? '"' : '\'';
                Putc(quotetype);
                DumpHTML(s, quotetype);
                Putc(quotetype);
            }
            else
                DumpHTML(s);
        }
        set<wstring>::const_iterator j;
        for(j=tag.Keys.begin(); j!=tag.Keys.end(); ++j)
        {
            Putc(' ');
            Dump(*j);
        }
        /* FIXME: Comments are moved to end of tag
                  (Probably nobody ever notices it)
         */
        list<wstring>::const_iterator k;
        for(k=tag.Comments.begin(); k!=tag.Comments.end(); ++k)
        {
            Putc('-'); Putc('-');
            Dump(*k);
            Putc('-'); Putc('-');
        }
        Putc('>');
    }
    void DumpRaw(const wstring &data) const
    {
        bool needcomments = false;
        for(unsigned a=0; a<data.size(); ++a)
        {
            if(data[a] == ' ' || data[a] == '\n'
            || data[a] == '\t' || data[a] == '\r')continue;
            needcomments = !IsEqual(data.substr(a, 4), "<!--");
            break;
        }
        if(needcomments)
        {
            FixedStyleScript++;
        }
        if(needcomments)
            Putc('<'), Putc('!'), Putc('-'), Putc('-'), Putc('\n');
        Dump(data);
        if(needcomments)
            Putc('\n'), Putc('-'), Putc('-'), Putc('>');
    }

    class Element : public ptrable
    {
    public:
        Element() {}
        virtual ~Element() {}
    };

    class ElemBody : public Element
    {
    public:
        wstring body;
        ElemBody(const wstring &b) : body(b)
        {
        }
    };

    class ElemTag : public Element
    {
    public:
        struct Tag tag;
        ElemTag(const struct Tag &t) : tag(t)
        {
        }
    };

    class ElemRaw : public Element
    {
    public:
        wstring data;
        ElemRaw(const wstring &b) : data(b)
        {
        }
    };

    vector<autoptr<Element> > Structure;

    void push_back(Element *e)
    {
        Structure.push_back(e);
    }

public:
    void Dump() const;
    
    void SetOut(const char *midset);

    void FilterText(wstring (*proc)(const wstring &));
    void ParseUCS4(FILE *fp);
    void Parse(FILE *fp, const char *charset = "iso-8859-1");
    
    Page(): underquote(0) {}
};

wstring Page::htmlencode(const wstring &s) const
{
    wstring res;
    unsigned a, b;
    for(a=b=0; a<s.size(); ++a)
    {
        if(s[a]=='<' && !underquote) /* Must be encoded: Could start a tag otherwise. */
        {
            if(a>b)res += s.substr(b, a-b);
            res += "&lt;";
            b = a+1; continue;
        }
        if(s[a]=='"' && underquote == '"') /* Must be encoded: Could end a parameter otherwise. */
        {
            if(a>b)res += s.substr(b, a-b);
            res += "&quot;";
            b = a+1; continue;
        }
        if(s[a]=='\'' && underquote == '\'') /* Must be encoded: Could end a parameter otherwise. */
        {
            if(a>b)res += s.substr(b, a-b);
            res += "&#39;";
            b = a+1; continue;
        }
        if(s[a]=='&') /* Must be encoded: Could start an entity otherwise. */
        {
            if(a>b)res += s.substr(b, a-b);
            res += "&amp;";
            b = a+1; continue;
        }
        if(s[a]==160) /* Not necessary, but nice: &nbsp; is commonly known. */
        {
    		/* FIXME: invent a better condition here */
        	if(!strict)
        	{
	            if(a>b)res += s.substr(b, a-b);
	            res += "&nbsp;";
	            b = a+1; continue;
	        }
        }
        if(lossless && !CanDump(s[a]))
        {
            char Buf[64];
            if(usehex)
                sprintf(Buf, "&#x%X;", s[a]);
            else
                sprintf(Buf, "&#%u;", s[a]);
            if(a>b)res += s.substr(b, a-b);
            res += Buf;
            b = a+1; continue;
        }
    }
    if(a>b)res += s.substr(b, a-b);
    return res;
}

wstring Page::htmldecode(const wstring &s) const
{
    wstring res;
    unsigned a, b;
    for(a=b=0; a<s.size(); )
    {
        if(s[a] != '&')
        {
            ++a;
            continue;
        }
        ucs4 specialchar = 0;
        unsigned c=a+1, e=0;
        if(c < s.size() && s[c] == '#')
        {
            ++c;
            if(c < s.size() && s[c] == 'x')
            {
                for(e=2; ++c < s.size(); ++e)
                {
                    if(s[c] >= '0' && s[c] <= '9') specialchar = specialchar*16+ (s[c]-'0');
                    else if(s[c] >= 'A' && s[c] <= 'F') specialchar = specialchar*16+ (s[c]-'A'+10);
                    else if(s[c] >= 'a' && s[c] <= 'f') specialchar = specialchar*16+ (s[c]-'a'+10);
                    else break;
                }
                goto AddChar;
            }
            for(e=1; c < s.size(); ++e, ++c)
            {
                if(s[c] >= '0' && s[c] <= '9') specialchar = specialchar*10+ (s[c]-'0');
                else break;
            }
            goto AddChar;
        }
        for(e=0; (c < s.size() && s[c]!=';'); ++e, ++c)
        {
            if(s[c] >= '0' && s[c] <= '9') { if(!e)break; continue; }
            if(s[c] >= 'A' && s[c] <= 'Z')continue;
            if(s[c] >= 'a' && s[c] <= 'z')continue;
            if(s[c] == '.')continue;
            break;
        }
        {wstring entname = s.substr(a+1, e);
         specialchar = FindEntity(entname);
        }
        if(specialchar != ilseq)
        {
AddChar:    res += s.substr(b, a-b);
            res += specialchar;
            a += e+1;
            if(a == s.size())break;
            if(s[a] == ';')
                ++a;
            else
            {
                /* FIXME: What if &abc; -code did not have ';' ? */
            }
            b = a;
            continue;
        }
        /* Unrecognized &..; -code. Don't parse it. */
        ++a;
        continue;
    }
    if(a > b) res += s.substr(b, a-b);
    return res;
}


void Page::Dump() const
{
    for(unsigned a=0; a<Structure.size(); ++a)
    {
        const Element *e = Structure[a];
        if(const ElemBody *elem = dynamic_cast<const ElemBody *> (e))
        {
            DumpHTML(elem->body);
        }
        else if(const ElemTag *elem = dynamic_cast<const ElemTag *> (e))
        {
            DumpTag(elem->tag);
        }
        else if(const ElemRaw *elem = dynamic_cast<const ElemRaw *> (e))
        {
        	DumpRaw(elem->data);
        }
    }
}

void Page::SetOut(const char *outset)
{
    Dumper.SetSet(outset);

    for(unsigned a=0; a<Structure.size(); ++a)
    {
        Element *e = Structure[a];
        if(ElemTag *elem = dynamic_cast<ElemTag *> (e))
        {
            struct Tag &tag = elem->tag;
            if(tag.Is("META"))
            {
                Tag::iterator i;
                i = tag.Ifind("HTTP-EQUIV");
                if(i == tag.end()) { continue; }
                if(!IsEqual(i->second, "CONTENT-TYPE"))
                {
                    continue;
                }
                
                i = tag.Ifind("CONTENT");
                if(i == tag.end())
                {
                    continue;
                }
                
                wstring &s = i->second;
                wstring tmp; tmp += "charset=";
                unsigned a = s.find(tmp);
                if(a == s.npos) { continue; }
                a += 8;
                
                tmp = s.substr(0, a);
                tmp += outset;
                i->second = tmp;
            }
        }
        else
        {
            /* We are only interested in meta-tags. */
        }
    }
}

void Page::FilterText(wstring (*proc)(const wstring &))
{
    for(unsigned a=0; a<Structure.size(); ++a)
    {
        Element *e = Structure[a];
        if(ElemBody *elem = dynamic_cast<ElemBody *> (e))
        {
            wstring &body = elem->body;
            body = proc(body);
        }
        else if(ElemTag *elem = dynamic_cast<ElemTag *> (e))
        {
            struct Tag &tag = elem->tag;
            if(tag.Is("IMG"))
            {
                Tag::iterator i;
                i = tag.Ifind("ALT");
                if(i != tag.end())
                {
                    i->second = proc(i->second);
                }
            }
            else if(tag.Is("A"))
            {
                Tag::iterator i;
                i = tag.Ifind("TITLE");
                if(i != tag.end())
                {
                    i->second = proc(i->second);
                }
            }
        }
        else
        {
            // Nothing interesting in ElemRaw
            // Are there other types of elements?
        }
    }
}
    
static void ParseInConv(FILE *fp, const char *inset, int fd)
{
	/* Deallocate all possible unnecessary resources */
    if(fileno(fp) != 0 && fd != 0) close(0);
    if(fileno(fp) != 1 && fd != 0) close(1);
    chdir("/");
    //signal(SIGHUP, _exit); - probably unsafe and redundant
    
    iconv_t converter = iconv_open(midset, inset);
    if(converter == (iconv_t)(-1))
    {
        perror("iconv_open");
        _exit(1);
    }
    
    char InBuf[4096];
    char OutBuf[4096];
    
    char *bufptr = InBuf;
    size_t bytes = 0;
    
    for(;;)
    {
        int code = fread(bufptr+bytes, 1, sizeof InBuf - bytes, fp);
        if(code <= 0)
        {
            if(!bytes)break;
        }
        else
            bytes += code;

    ReCode:
        char *outptr = OutBuf;
        size_t outsize = sizeof OutBuf;
        
        bool needspace = false;
        bool gotilseq = false;
    #if DEBUG
        fprintf(stderr, "P2:Converting %u bytes to %u bytes space\n", bytes, outsize);
        unsigned bytesread = bytes;
    #endif
        unsigned converted = outsize;
        size_t retval = iconv(converter, &bufptr, &bytes, &outptr, &outsize);
        converted -= outsize;
    #if DEBUG
        bytesread -= bytes;
    #endif
        write(fd, OutBuf, converted);
        
        if(retval == (size_t)-1)
        {
    #if DEBUG
            perror("iconv");
    #endif
            if(errno == E2BIG)
            {
                needspace = true;
            }
            if(errno == EILSEQ)
            {
                gotilseq = true;
            }
            if(errno == EINVAL)
            {
                /* Got partial byte */
            }
        }
    #if DEBUG
        fprintf(stderr, "%u bytes read, %u bytes left, %u bytes generated, %u bytes space left\n", bytesread, bytes, converted, outsize);
        fflush(stderr);
    #endif
        if(gotilseq)
        {
            write(fd, &ilseq, sizeof(ilseq));
            /* Skip the invalid byte */
            --bytes; ++bufptr;
        }
        
        if(needspace)
        {
            /* No need to retry reading, just want more space */
            goto ReCode;
        }
        
        memmove(&InBuf[0], bufptr, bytes);
        bufptr = InBuf;
    }
    
    iconv_close(converter);
}

void Page::Parse(FILE *fp, const char *charset)
{
    string inset = charset;
ReHandle:
    if(verbose >= 1)
    {
        fprintf(stderr, "Recoding input (%s) to %s\n", inset.c_str(), midset);
        fflush(stderr);
    }
    int pip[2]; pipe(pip);
    int pid = fork();
    if(!pid)
    {
        close(pip[0]);
    	ParseInConv(fp, inset.c_str(), pip[1]);
        _exit(0);
    }
    close(pip[1]);
    FILE *pipfp = fdopen(pip[0], "rb");
    
    Structure.clear();
    ParseUCS4(pipfp);
    
    fclose(pipfp);
    kill(pid, SIGHUP);
    waitpid(pid, NULL, 0);

    for(unsigned a=0; a<Structure.size(); ++a)
    {
        const Element *e = Structure[a];
        if(const ElemTag *elem = dynamic_cast<const ElemTag *> (e))
        {
            const struct Tag &tag = elem->tag;
            if(tag.Is("META"))
            {
                Tag::const_iterator i;
                i = tag.Ifind("HTTP-EQUIV");
                if(i == tag.end()) { continue; }
                if(!IsEqual(i->second, "CONTENT-TYPE"))
                {
                    continue;
                }
                i = tag.Ifind("CONTENT");
                if(i == tag.end())
                {
                    continue;
                }
                
                const wstring &s = i->second;
                wstring tmp; tmp += "charset=";
                unsigned a = s.find(tmp);
                if(a == s.npos) { continue; }
                a += 8;
                wstring newcharset = s.substr(a);
                if(!IsEqual(newcharset, inset.c_str()))
                {
                    inset = "";
                    for(unsigned a=0; a<newcharset.size(); ++a)
                        inset += (unsigned char)newcharset[a];
                    
                    if(verbose >= 0)
                    {
                        fprintf(stderr,
                            "Warning: Document character encoding seems to be %s, which differs from what you specified. Rereading.\n",
                            inset.c_str());
                    }
                    
                    if(fseek(fp, 0, SEEK_SET) == -1)
                    {
                        if(errno == ESPIPE
                        || errno == EBADF)
                        {
                            fprintf(stderr, "Error: stdin is not seekable. Can not reread! You should use the -I%s option. Or perhaps you committed a Useless Use Of Cat.\n",
                                inset.c_str());
                            return;
                        }
                        else
                            perror("fseek");
                    }
                   
                    goto ReHandle;
                }
            }
        }
        else
        {
            /* We are only interested in meta-tags. */
        }
    }
}

void Page::ParseUCS4(FILE *fp)
{
    wstring body, comment, rawcontent;
    
    enum states
    {
        stBody,
        stTagName,
        stTagSpace,
        stTagSGML,
        stTagSGMLquoted,
        stTagSGMLquoted2,
        stTagParam,
        stTagParamValue,
        stTagParamValueQuoted,
        stTagParamValueQuoted2,
#if 1
        stRawContent,
        stMaybeEndRawContent,
        stRawContentTagName,
#endif
        stMaybeComment,
        stComment,
        stMaybeEndComment
    } state = stBody;
    
    wstring ParamName;
    wstring ParamValue;
    wstring RawCurrentTag;
    Tag Tag;
    
    for(;;)
    {
        int c = Getc(fp);
        if(feof(fp))break;
        
Statechange:
        //fprintf(stderr, "State=%d, c=%c\n", state, c);
        switch(state)
        {
            case stBody:
                if(c != '<')
                {
                    body += c;
                    break;
                }
                if(body.size())
                {
                    push_back(new ElemBody(htmldecode(body)));
                    CLEARSTR(body);
                }
                
                CLEARSTR(Tag.Name);
                state = stTagName;
                break;

            case stTagName:
                if((c == '!' && !Tag.Name.size())
                || (c == '/' && !Tag.Name.size())
                || (c >= 'A' && c <= 'Z')
                || (c >= 'a' && c <= 'z')
                || (c >= '0' && c <= '9' && Tag.Name.size())
                || (c == '_' || c == '.'
                 || c == ':') /* No - here, it breaks comments */
                  )
                {
                    Tag.Name += c;
                    break;
                }
                CLEARSTR(ParamName);
                state = stTagSpace;
                goto Statechange;
                
            case stTagSpace:
                if(c == '-')
                {
                    state = stMaybeComment;
                    break;
                }
                if(c == '>')
                {
                    if(ParamName.size())
                    {
                        Tag.SetParam(ParamName, wstring());
                        CLEARSTR(ParamName);
                    }
                    push_back(new ElemTag(Tag));
                    
                    state = stBody;
                    
                    if(!strict)
                    {
                        if(Tag.Is("SCRIPT")
                        || Tag.Is("STYLE"))
                        {
                            RawCurrentTag = Tag.Name;
                            RawCurrentTag.insert(0, 1, (ucs4)'/');
                            state = stRawContent;
                        }
                    }

                    Tag.clear();
                    
                    CLEARSTR(body);
                    break;
                }
                if(Tag.Name.size() && Tag.Name[0] == '!')
                {
                    if(c != ' ' && c != '\t' && c != '\v' && c != '\r' && c != '\n')
                    {
                        if(ParamName.size())
                            ParamName += ' ';
                        state = stTagSGML;
                        goto Statechange;
                    }
                }
                
                if((c >= 'A' && c <= 'Z')
                || (c >= 'a' && c <= 'z'))
                {
                    state = stTagParam;
                    goto Statechange;
                }

                /* TODO: Handle unrecognized character in tag */
                break;
            
            case stTagSGML:
                if(c == '-')
                {
                    state = stMaybeComment;
                    break;
                }
                if(c == '>')
                {
                    state = stTagSpace;
                    goto Statechange;
                }
                ParamName += c;
                if(c == '"')
                    state = stTagSGMLquoted;
                else if(c == '\'')
                    state = stTagSGMLquoted2;
                break;
            
            case stTagSGMLquoted:
                ParamName += c;
                if(c == '"')
                    state = stTagSGML;
                break;

            case stTagSGMLquoted2:
                ParamName += c;
                if(c == '\'')
                    state = stTagSGML;
                break;

            case stTagParam:
                if((c >= 'A' && c <= 'Z')
                || (c >= 'a' && c <= 'z')
                || (c == '-')  /* apparently http-equiv is a valid name for param */
                  )
                {
                    ParamName += c;
                    break;
                }

                if(c == '=')
                {
                    CLEARSTR(ParamValue);
                    state = stTagParamValue;
                    break;
                }

                Tag.SetParam(ParamName);
                CLEARSTR(ParamName);
                state = stTagSpace;
                goto Statechange;
            
            case stTagParamValue:
                if(!ParamValue.size())
                {
                    if(c == '"')
                    {
                        state = stTagParamValueQuoted;
                        break;
                    }
                    if(c == '\'')
                    {
                        state = stTagParamValueQuoted2;
                        break;
                    }
                }
                if((c >= 'A' && c <= 'Z')
                || (c >= 'a' && c <= 'z')
                || (c >= '0' && c <= '9')
                || c == '.' || c == '-' /* .-_: are allowed by sgml */
                || c == '_' || c == ':'
                || (!strict && (c == ',' || c == '/' || c == '%' || c == '#' || c == '~')
                   ) /* generated by many programs :-/ */
                  )
                {
                    if(c == ',' || c == '/' || c == '%' || c == '#' || c == '~')
                        ParamViolations++;
                    ParamValue += c;
                    break;
                }
                Tag.SetParam(ParamName, htmldecode(ParamValue));
                CLEARSTR(ParamName);
                CLEARSTR(ParamValue);
                state = stTagSpace;
                goto Statechange;
            
            case stTagParamValueQuoted:
                if(c == '"')
                {
                    Tag.SetParam(ParamName, htmldecode(ParamValue));
                    CLEARSTR(ParamName);
                    CLEARSTR(ParamValue);
                    state = stTagSpace;
                    break;
                }
                ParamValue += c;
                break;

            case stTagParamValueQuoted2:
                if(c == '\'')
                {
                    Tag.SetParam(ParamName, htmldecode(ParamValue));
                    CLEARSTR(ParamName);
                    CLEARSTR(ParamValue);
                    state = stTagSpace;
                    break;
                }
                ParamValue += c;
                break;

            case stMaybeComment:
                if(c == '-')
                {
                    state = stComment;
                    CLEARSTR(comment);
                    break;
                }
                
                ParamName += '-';
                
                state = stTagSpace;
                goto Statechange;
            
            case stComment:
                if(c == '-')
                {
                    state = stMaybeEndComment;
                    break;
                }
                comment += c;
                break;
            
            case stMaybeEndComment:
                if(c == '-')
                {
                    if(comment.size())
                    {
                        Tag.Comments.push_back(comment);
                        CLEARSTR(comment);
                    }
                    
                    state = stTagSpace;
                    break;
                }
                comment += '-';
                state = stComment;
                goto Statechange;
#if 1
            case stRawContent:
                // end with </
                if(c == '<')
                {
                    state = stMaybeEndRawContent;
                    break;
                }
                rawcontent += c;
                break;
            
            case stMaybeEndRawContent:
                if(c == '/')
                {
                    CLEARSTR(Tag.Name);
                    state = stRawContentTagName;
                    goto Statechange;
                }
                rawcontent += '<';
                state = stRawContent;
                goto Statechange;

            case stRawContentTagName:
                if((c == '!' && !Tag.Name.size())
                || (c == '/' && !Tag.Name.size())
                || (c >= 'A' && c <= 'Z')
                || (c >= 'a' && c <= 'z')
                || (c >= '0' && c <= '9' && Tag.Name.size())
                || (c == '_' || c == '.'
                 || c == ':') /* No - here, it breaks comments */
                  )
                {
                    Tag.Name += c;
                    break;
                }
                
                if(Tag.Is(RawCurrentTag))
                {
                    if(rawcontent.size())
                    {
                        push_back(new ElemRaw(rawcontent));
                        CLEARSTR(rawcontent);
                    }
                    
                    CLEARSTR(ParamName);
                    state = stTagSpace;
                    goto Statechange;
                }
                rawcontent += '<';
                rawcontent += Tag.Name;
                state = stRawContent;
                goto Statechange;
#endif
        }
    }
    if(body.size())
    {
        push_back(new ElemBody(htmldecode(body)));
        CLEARSTR(body);
    }
    if(rawcontent.size())
    {
        push_back(new ElemRaw(rawcontent));
        CLEARSTR(rawcontent);
    }
    /* Don't add a broken tag to the end. */
}

#if 0
/* This is a support for altering the text content */
static wstring testifiltteri(const wstring &var)
{
    wstring res;
    // res += '[';
    res += var;
    // res += ']';
    return res;
}
#endif

#include <argh.hh>

int main(int argc, const char *const *argv)
{
    string outset = "iso-8859-1";
    string inset = "iso-8859-1";
    
    ParamHandler Argh;
    Argh.AddLong("inset",   'I').SetString().SetDesc("Assumed input character set (default: "+inset+")", "setname");
    Argh.AddLong("outset",  'O').SetString().SetDesc("Wanted output character set (default: "+outset+")", "setname");
    Argh.AddLong("help",    'h').SetBool().SetDesc("This help.");
    Argh.AddLong("lossy",   'l').SetBool().SetDesc("Disable lossless conversion.");
    Argh.AddLong("usehex",  'x').SetBool().SetDesc("Use hexadecimal escapes.");
    Argh.AddLong("version", 'V').SetBool().SetDesc("Displays version information.");
    Argh.AddLong("strict",  's').SetBool().SetDesc("Turn off support for slightly broken HTML.");
    Argh.AddLong("verbose", 'v').SetBool().SetDesc("Be less quiet.");
    Argh.AddLong("quiet",   'q').SetBool().SetDesc("Be less verbose.");

    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'V': printf("%s\n", VERSION); return 0;
            case 'I': inset = Argh.GetString(); break;
            case 'O': outset = Argh.GetString(); break;
            case 'l': lossless = !Argh.GetBool(); break;
            case 'x': usehex = Argh.GetBool(); break;
            case 's': strict = Argh.GetBool(); break;
            case 'v': verbose += Argh.GetBool() ? 1 : -1; break;
            case 'q': verbose -= Argh.GetBool() ? 1 : -1; break;
            case 'h':
                printf(
                    "htmlrecode " VERSION " - Copyright (C) 1992,2003 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\n"
                    "Usage: htmlrecode [<option> [<...>]]\n"
                    "\n"
                    "Reads stdin, writes stdout.\n"
                    "\nOptions:\n");
                Argh.ListOptions();
                printf("\n"
                    "Pipe in the html file and pipe the output to result file.\n");
                return 0;
            default:
                // TODO
                break;
        }
    }
    if(!Argh.ok())return -1;

    Page p;

	rewind(stdin);
    p.Parse(stdin, inset.c_str());
    fclose(stdin);
    
    /* This is a support for altering the text content */
    //p.FilterText(testifiltteri);
    
    p.SetOut(outset.c_str());
    p.Dump();
    
    if(FixedStyleScript && verbose >= 1)
    {
        fprintf(stderr,
            "Warning: Fixed %u SCRIPT/STYLE block%s that %sn't properly hidden with HTML comments. Be grateful.\n",
                FixedStyleScript,
                FixedStyleScript==1 ? "" : "s",
                FixedStyleScript==1 ? "was" : "were");
    }
    if(ParamViolations && verbose >= 1)
    {
        fprintf(stderr,
            "Warning: Fixed %u broken (incorrectly unquoted) tag parameter%s. Be grateful.\n",
                ParamViolations,
                ParamViolations==1 ? "" : "s"
               );
    }
    
    return 0;
}
